/*
 * Copyright 2003 Jukka Pihl <jukka dot pihl ät iki dot fi>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Jukka Pihl
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * memio.c  - memory mapped io for unzip-library
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <assert.h>


#include "zlib.h"
#include "ioapi.h"
#include "unzip.h"

#define M_MAGIC 238764

typedef struct memstream 
{
    int   len;
    int   pos;
    unsigned char *ptr;
    unsigned char *ep;
    int   magic;
} memstream;


static voidpf ZCALLBACK mopen_file_func (voidpf opaque, const char* filename, int mode)
{
    const char* mode_fopen = NULL;
    int prot = 0;
    int oflags = 0;
    void *ptr;
    int fd;
    int y,i;
    int priv = 0;
    struct stat st;
    memstream *m;
    

    if ((mode & ZLIB_FILEFUNC_MODE_READWRITEFILTER)==ZLIB_FILEFUNC_MODE_READ) {
        mode_fopen = "rb";
	prot  = PROT_READ;
	oflags = O_RDONLY;
	priv = 1;
    } else  if (mode & ZLIB_FILEFUNC_MODE_EXISTING) {
        mode_fopen = "r+b";
	prot  = PROT_READ | PROT_WRITE;
	oflags = O_RDWR;
    } else if (mode & ZLIB_FILEFUNC_MODE_CREATE) {
        mode_fopen = "wb";
	prot  = PROT_READ | PROT_WRITE;
	oflags = O_RDWR | O_CREAT;
    } else {
	prot  = PROT_READ;
	oflags = O_RDONLY;
    }

    fd = open(filename, oflags, 0666);
    if(fd < 0)
	return NULL;

    if(fstat(fd, &st) < 0) {
	close(fd);
	return NULL;
    }

    ptr = mmap(NULL, st.st_size, prot, priv ? MAP_PRIVATE : MAP_SHARED, fd, 0);

    if((ptr == MAP_FAILED) || (ptr == NULL)) {
	close(fd);
	return NULL;
    }

    close(fd);

    m = calloc(sizeof(*m),1);
    m->magic = M_MAGIC;
    m->ptr   = ptr;
    m->pos   = 0;
    m->len   = st.st_size;
    m->ep    = m->ptr + m->len;

    y = 1;
    for(i=0; i < m->len; i++) y ^= m->ptr[i];

    return (void*) m;
}


static uLong ZCALLBACK mread_file_func (voidpf opaque, voidpf stream, void* buf, uLong size)
{
    int left;
    memstream *m = (memstream *) stream;
    assert(m->magic == M_MAGIC);
    assert(m->pos >= 0);
    assert(m->pos <= m->len);
    assert(size > 0);
    assert(size < 100000000);
    
    left = m->len - m->pos;

    if(left <= 0)
	return 0;

    if(left < size)
	size = left;

    memcpy(buf, m->ptr + m->pos, size);

    m->pos += size;
    assert(m->pos > 0);
    assert(m->pos <= m->len);

    return size;
}

static uLong ZCALLBACK mwrite_file_func (voidpf opaque, voidpf stream, const void* buf, uLong size)
{
    // no write implemented!!!!!
    return size;
}

static long ZCALLBACK mtell_file_func (voidpf opaque, voidpf stream)
{
    memstream *m = (memstream *) stream;
    assert(m->magic == M_MAGIC);
    return m->pos;
}

static long ZCALLBACK mseek_file_func (voidpf opaque, voidpf stream, long offset, int origin)
{
    memstream *m = (memstream *) stream;
    assert(m->magic == M_MAGIC);

    switch (origin)
    {
    case ZLIB_FILEFUNC_SEEK_CUR :
	m->pos += offset;
        break;
    case ZLIB_FILEFUNC_SEEK_END :
	m->pos = m->len + offset;
        break;
    case ZLIB_FILEFUNC_SEEK_SET :
	m->pos = offset;
        break;
    default: 
	return -1;
    }
    return 0;
}

static int ZCALLBACK mclose_file_func (voidpf opaque, voidpf stream)
{
    memstream *m = (memstream *) stream;
    assert(m->magic == M_MAGIC);

    if(m->ptr) {
	munmap(m->ptr, m->len);
	m->ptr = NULL;
    }
    m->magic = 0;
    free(m);
    return 0;

}

static int ZCALLBACK merror_file_func (voidpf opaque, voidpf stream)
{
    memstream *m = (memstream *) stream;
    assert(m->magic == M_MAGIC);
    return 0;
}

void fill_mopen_filefunc (zlib_filefunc_def* pzlib_filefunc_def)
{
    pzlib_filefunc_def->zopen_file = mopen_file_func;
    pzlib_filefunc_def->zread_file = mread_file_func;
    pzlib_filefunc_def->zwrite_file = mwrite_file_func;
    pzlib_filefunc_def->ztell_file = mtell_file_func;
    pzlib_filefunc_def->zseek_file = mseek_file_func;
    pzlib_filefunc_def->zclose_file = mclose_file_func;
    pzlib_filefunc_def->zerror_file = merror_file_func;
    pzlib_filefunc_def->opaque = NULL;
}


zlib_filefunc_def memio = {
    mopen_file_func,
    mread_file_func,
    mwrite_file_func,
    mtell_file_func,
    mseek_file_func,
    mclose_file_func,
    merror_file_func,
    NULL
};
