/*
 * Copyright 2003 Jukka Pihl <jukka dot pihl ät iki dot fi>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Jukka Pihl
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* 
**  mod_ziplook.c -- Apache ziplook module
*/ 

    /* The section for the Configure script:
     * MODULE-DEFINITION-START
     * Name: ziplook_module
     * ConfigStart
    echo "      enabling zlib support for mod_rewrite"
    LIBS="$LIBS -lz"
     * ConfigEnd
     * MODULE-DEFINITION-END
     */


#define CORE_PRIVATE
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "util_script.h"
#include "http_config.h"
#include <time.h>
#include "http_log.h"


#include "apr_pools.h"
#include "apr_strings.h"

#if APR_HAVE_UNISTD_H
#include <unistd.h>
#endif
#define APR_WANT_STRFUNC
#include "apr_want.h"

#ifdef unix
# include <unistd.h>
# include <utime.h>
#else
# include <direct.h>
# include <io.h>
#endif

#include "unzip.h"


#define ZIPLOOK_VERSION_STR  "0.99"

#include "ioapi.c"
#include "memio.c"
#include "unzip.c"



// #define HAVE_CRC  
#define HAVE_NAME 


#define CASESENSITIVITY (1)
#define WRITEBUFFERSIZE (8192)
#define MAXFILENAME (512)

#define NORMAL  0 
#define PLAIN   1
#define LINKED  2
#define SLINKED 3
#define NONE    4


typedef struct zipdata {
    request_rec   *r;
    const char    *filename;
    char          *path;
    char          *file;
    apr_table_t   *query;
    unz_file_info *file_info;
    unzFile        uf;
    int            accept_gzip;
} zipdata;


AP_DECLARE(void) ap_block_alarms(void) { }
AP_DECLARE(void) ap_unblock_alarms(void) { }


static apr_table_t *get_query(request_rec *r)
{
    char *str,*s1,*s2;
 
    apr_table_t *nq = NULL;

    str = r->args;

    if(!str)
	return NULL;

    while(str && *str) {
	
	s1 = ap_getword_nc(r->pool,&str,'&');
	
	if(!s1) continue;
	
	if((s2 = strchr(s1,'=')) != NULL) {
	    *s2++ = 0;
	} else {
	    s2 = NULL;
	}
	    
	// anything to unescape
	    
	if(strchr(s1,'%') || strchr(s1,'+') ) {
	    s1 = apr_pstrdup(r->pool, s1);
	    ap_unescape_url(s1);
	}

	if(s2) {
	    if(strchr(s2,'%') || strchr(s2,'+') ) {
		s2 = apr_pstrdup(r->pool, s2);
		ap_unescape_url(s2);
	    } 
	}


	if(nq == NULL) nq = apr_table_make(r->pool, 16);

	apr_table_add(nq, s1, s2);
    }

    return nq;
}


static char comment[1001];

static const char *get_string_method(unz_file_info *file_info)
{
    const char *string_method;

    if (file_info->compression_method==0) {
	string_method="Stored";
    } else if (file_info->compression_method==Z_DEFLATED) {
	uInt iLevel=(uInt)((file_info->flag & 0x6)/2);
	if (iLevel==0)
	    string_method="Defl:N";
	else if (iLevel==1)
	    string_method="Defl:X";
	else if ((iLevel==2) || (iLevel==3))
	    string_method="Defl:F"; /* 2:fast , 3 : extra fast*/
	else 
	    string_method="Unkn. ";
    } else {
	string_method="Unkn. ";
    }
    return string_method;
}

#define XFL_ZERO 0
#define XFL_MAX  2
#define XFL_FAST 4

static int get_xfl(unz_file_info *file_info)
{
    if (file_info->compression_method==0)
	return XFL_ZERO;

    if (file_info->compression_method==Z_DEFLATED) {
	uInt iLevel=(uInt)((file_info->flag & 0x6)/2);
	if (iLevel==0)
	    return XFL_ZERO;

	else if (iLevel==1)
	    return XFL_MAX;
	else if ((iLevel==2) || (iLevel==3))
	    return XFL_FAST;    /* 2:fast , 3 : extra fast*/
	else 
	    return XFL_ZERO;
    }

    
    return XFL_ZERO;
}


/*
  zip-versions:

    0 - MS-DOS and OS/2 (FAT / VFAT / FAT32 file systems)
    1 - Amiga                     
    2 - VAX/VMS
    3 - Unix                      
    4 - VM/CMS
    5 - Atari ST                  
    6 - OS/2 H.P.F.S.
    7 - Macintosh                 
    8 - Z-System
    9 - CP/M                     
    10 - Windows NTFS
    11 thru 255 - unused


  gzip-verisions:
    0 - FAT filesystem (MS-DOS, OS/2, NT/Win32) 1 - Amiga 
    2 - VMS (or OpenVMS) 
    3 - Unix 
    4 - VM/CMS 
    5 - Atari TOS 
    6 - HPFS filesystem (OS/2, NT) 
    7 - Macintosh 
    8 - Z-System 
    9 - CP/M 
    10 - TOPS-20 
    11 - NTFS filesystem (NT) 
    12 - QDOS 
    13 - Acorn RISCOS 
    255 - unknown 

*/



static int get_version(unz_file_info *file_info)
{
    return  (file_info->version);

}


static int do_list(zipdata *zd)
{
    uLong i;
    unz_global_info gi;
    int err;
    int style = LINKED;
    long long total_u_size = 0;
    long long total_c_size = 0;
    int total_ratio = 0;
    request_rec *r = zd->r;
    unzFile uf     = zd->uf;
    apr_table_t *q       = zd->query;

    if(q) {
	const char *st;
	if(((st = apr_table_get(q,"style")))) {
	    if(!strcmp(st,"normal")) {
		style = NORMAL;
	    } else if(!strcmp(st,"plain")) {
		style = PLAIN;
	    } else if(!strcmp(st,"linked")  || !strcmp(st,"long")) { 
		style = LINKED;
	    } else if(!strcmp(st,"slinked") || !strcmp(st,"short")) {
		style = SLINKED;
	    } else if(!strcmp(st,"none")) {
		style = NONE;
	    } else {
		style = SLINKED;
	    }
	}
    }

    ap_update_mtime(r,r->finfo.mtime);

    ap_set_last_modified(r);
    if ((err = ap_meets_conditions(r)) != OK) {
	r->status = err;
	ap_set_last_modified(r);
	// ap_send_http_header(r);
	return err;
    }

    if (r->header_only) {
	r->content_type = "text/html";      
	// ap_send_http_header(r);
	return OK;
    }


    r->content_type = "text/html"; 
    // ap_send_http_header(r);

    
    err = unzGetGlobalInfo (uf,&gi);
    if (err!=UNZ_OK) {
	ap_rprintf(r,"(ziplook) error %d with zipfile in unzGetGlobalInfo \n",err);
	return OK;
    }

    
    // memset(comment,0,sizeof(comment));
    // unzGetGlobalComment (uf, comment,1000);

 
    ap_rprintf(r,"<body bgcolor=\"#ffffff\">\n");

    ap_rprintf(r,"<a href=\"%s?style=plain\">plain</a>\n", r->uri);
    ap_rprintf(r,"&nbsp;<a href=\"%s?style=long\">long</a>\n", r->uri);
    ap_rprintf(r,"&nbsp;<a href=\"%s?style=short\">short</a>\n", r->uri);
    ap_rprintf(r,"&nbsp;<a href=\"%s?style=none\">none</a>\n", r->uri);
    ap_rprintf(r,"&nbsp;<a href=\"%s?fetch=1\">fetch</a>\n", r->uri);

    ap_rprintf(r,"<br>\n");

    // if(comment[0]) { ap_rprintf(r,"zip comment: %s\n<br>\n", comment);  }

    ap_rprintf(r,"<pre>\n");

    switch(style) {
    case NONE:
	break;
    case NORMAL:
    case LINKED:
    default:
	ap_rprintf(r,"<pre>\n");
	ap_rprintf(r," Length  Method   Size  Ratio   Date    Time   CRC-32     Name\n");
	ap_rprintf(r," ------  ------   ----  -----   ----    ----   ------     ----\n");
	break;
    case SLINKED:
    case PLAIN:
	break;
    }

    for (i=0;i<gi.number_entry;i++)
    {
	char filename_inzip[256+1];
	unz_file_info file_info;
	uLong ratio=0;
	const char *string_method;
	err = unzGetCurrentFileInfo(uf,&file_info,
				    filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);
	if (err!=UNZ_OK)
	{
	    ap_rprintf(r,"error %d with zipfile in unzGetCurrentFileInfo\n",err);
	    break;
	}

	if (file_info.uncompressed_size>0)
	    ratio = (file_info.compressed_size*100)/file_info.uncompressed_size;

	string_method = get_string_method(&file_info);

	total_u_size += file_info.uncompressed_size;
	total_c_size += file_info.compressed_size;

	switch(style) {
	default:
	case NORMAL:
	    ap_rprintf(r,
		       "%7lu  %6s %7lu %3lu%%  %2.2lu-%2.2lu-%2.2lu  %2.2lu:%2.2lu  %8.8lx   %s\n",
		       file_info.uncompressed_size,string_method,file_info.compressed_size,
		       ratio,
		       (uLong)file_info.tmu_date.tm_mon + 1,
		       (uLong)file_info.tmu_date.tm_mday,
		       (uLong)file_info.tmu_date.tm_year % 100,
		       (uLong)file_info.tmu_date.tm_hour,(uLong)file_info.tmu_date.tm_min,
		       (uLong)file_info.crc,filename_inzip);
	    break;
	case LINKED:
	    ap_rprintf(r,
		       "%7lu  %6s %7lu %3lu%%  %2.2lu-%2.2lu-%2.2lu  %2.2lu:%2.2lu  %8.8lx",
		       file_info.uncompressed_size,string_method,file_info.compressed_size,
		       ratio,
		       (uLong)file_info.tmu_date.tm_mon + 1,
		       (uLong)file_info.tmu_date.tm_mday,
		       (uLong)file_info.tmu_date.tm_year % 100,
		       (uLong)file_info.tmu_date.tm_hour,(uLong)file_info.tmu_date.tm_min,
		       (uLong)file_info.crc);

	    ap_rprintf(r,"   <a href=\"%s/%s\">%s</a>\n", r->uri, filename_inzip, filename_inzip);
	    break;

	case SLINKED:
	    ap_rprintf(r,"%8ld <a href=\"%s/%s\">%s</a>\n", 
		       file_info.uncompressed_size,
		       r->uri, filename_inzip, filename_inzip);
	    break;
	case PLAIN:
	    ap_rprintf(r,"%s\n", filename_inzip);
	    break;
	case NONE:
	    break;
	}

	if ((i+1)<gi.number_entry)
	{
	    err = unzGoToNextFile(uf);
	    if (err!=UNZ_OK)
	    {
		ap_rprintf(r,"error %d with zipfile in unzGoToNextFile\n",err);
		break;
	    }
	}
    }

    total_ratio = (total_c_size*100)/total_u_size;

    ap_rprintf(r,"\n");
    ap_rprintf(r,"total uncompressesed size: %8d\n", (int) total_u_size);
    ap_rprintf(r,"total   compressesed size: %8d\n", (int) total_c_size);
    ap_rprintf(r,"total               ratio: %8d\n", total_ratio);

    ap_rprintf(r,"</pre>\n");

    
    ap_rprintf(r,"</body>\n");

    return OK;
}

#if 0
static const char *get_content_type(request_rec *r, const char *filename)
{

    request_rec *re = ap_sub_req_lookup_file(filename, r, NULL);
    if(re->content_type)
	return re->content_type;
    
    return "text/plain";
}

#endif

#if 0

extern module mime_module;

static const char *get_content_type(request_rec *r, const char *filename)
{
    char *oldfile = r->filename;
    r->filename = (char *) filename;
    r->content_type = NULL;
    mime_module.type_checker(r);
    r->filename = oldfile;

    if(r->content_type)
	return r->content_type;
    
    return "text/plain";
}

#endif

static const char *get_content_type(request_rec *r, const char *filename)
{
    return r->content_type ? r->content_type : "text/plain";
}


static apr_time_t get_time(request_rec *r, unz_file_info *info)
{
    struct tm ti;

    memset(&ti,0,sizeof(ti));

    ti.tm_year = info->tmu_date.tm_year;
    ti.tm_mon  = info->tmu_date.tm_mon;
    ti.tm_mday = info->tmu_date.tm_mday;
    ti.tm_hour = info->tmu_date.tm_hour,
    ti.tm_min  = info->tmu_date.tm_min;
    ti.tm_sec  = info->tmu_date.tm_sec;

    if(ti.tm_year < 80) {
	ti.tm_year += 100;
    } else if(ti.tm_year >= 1900) {
	ti.tm_year -= 1900;
    }

    return apr_time_from_sec(mktime(&ti));
}

// -------------------------------------------------------------

// -------------------------------------------------------------


#define GZIP_FTEXT    0x01
#define GZIP_FHCRC    0x02
#define GZIP_FEXTRA   0x04
#define GZIP_FNAME    0x08
#define GZIP_FCOMMENT 0x10


#if defined(HAVE_CRC) && defined(HAVE_NAME)
#define MY_FLAGS (GZIP_FHCRC | GZIP_FNAME)
#elif defined(HAVE_CRC)
#define MY_FLAGS (GZIP_FHCRC)
#elif defined(HAVE_NAME)
#define MY_FLAGS (GZIP_FNAME)
#else
#define MY_FLAGS 0
#endif

// -------------------------------------------------------------


static int send_compressed(zipdata *zd)
{
    request_rec *r  = zd->r;
    unzFile      uf = zd->uf;
    unz_file_info *file_info = zd->file_info;
    int err=UNZ_OK;
    void* buf;
    uInt size_buf;
    int ratio=0;
    int method = 0;
    int level  = 0;
    apr_time_t mtime;
    time_t     unix_mtime;
    int i;
    int flen;
#ifdef HAVE_CRC
    unsigned int c16;
#endif
    unsigned char gtable[16];

    if (file_info->uncompressed_size>0)
	ratio = (file_info->compressed_size*100)/file_info->uncompressed_size;

    size_buf = WRITEBUFFERSIZE;
    buf = (void*) apr_palloc(r->pool, size_buf);
    if (buf==NULL)
    {
	r->content_type = "text/html";
	// ap_send_http_header(r);
        ap_rprintf(r,"Error allocating memory\n");

	apr_table_addn(r->notes,"ziplook", "sent-gzip (memory-allocation-error)");

        return UNZ_INTERNALERROR;
    }

    err = unzOpenCurrentFile2(uf, &method, &level, 1);

    if (err!=UNZ_OK)
    {
	r->content_type = "text/html"; 
	// ap_send_http_header(r);
	ap_rprintf(r,"error %d with zipfile in unzOpenCurrentFile\n",err);

	apr_table_addn(r->notes,"ziplook", "sent-gzip (cannot open)");

	return -1;
    }

    r->content_type = get_content_type(r,zd->file);

    apr_table_setn(r->headers_out,
		   "Zip-info",
		   apr_psprintf(r->pool, 
				"crc=%8.8lx, csize=%ld, usize=%ld, ratio=%d, method=%s", 
				file_info->crc,
				file_info->compressed_size,
				file_info->uncompressed_size,
				ratio,
				get_string_method(file_info)));
    
    mtime = get_time(r, file_info);
    unix_mtime = apr_time_sec(mtime);
    

    ap_update_mtime(r,mtime);

    if((err = ap_meets_conditions(r)) != OK) {
	r->status = err;
	ap_set_last_modified(r);
	// ap_send_http_header(r);

	apr_table_addn(r->notes,"ziplook", "sent-gzip (not-changed)");
	    
	return err;
    }

 
    apr_table_setn(r->headers_out, "Content-Encoding", "gzip");
    apr_table_setn(r->headers_out, "Vary", "Accept-Encoding");

    // ap_send_http_header(r);


#ifdef HAVE_CRC
    c16 = crc32(0, Z_NULL, 0);
#endif
    i=0;
    gtable[i++] = 31;   // ID1 (31)
    gtable[i++] = 139;  // ID2 (139)
    gtable[i++] = 8;    // Compression Method = deflate (8)
    gtable[i++] = MY_FLAGS; // flags: FNAME exits  1<<3; 
    gtable[i++] = (unix_mtime >> 0)  & 0xff;   // modification time
    gtable[i++] = (unix_mtime >> 8)  & 0xff;
    gtable[i++] = (unix_mtime >> 16) & 0xff;
    gtable[i++] = (unix_mtime >> 24) & 0xff;
    gtable[i++] = get_xfl(file_info);     // extra flags  
    gtable[i++] = get_version(file_info); // version


    ap_rwrite(gtable, i, r);
#ifdef HAVE_CRC
    c16 = crc32(c16, gtable, i);
#endif


#ifdef HAVE_NAME
    {
	char *f = zd->file;
	if(f[0] == '/' && f[1]) 
	    f++;
	flen = strlen(f)+1;
	ap_rwrite(f,flen,r);
#ifdef HAVE_CRC
	c16 = crc32(c16, f,flen);
#endif
    }
#endif

#ifdef HAVE_CRC
    i=0; 
    gtable[i++] = (c16 >> 0) & 0xff;
    gtable[i++] = (c16 >> 8) & 0xff;
    ap_rwrite(gtable, 2,r);
#endif

    do {
	err = unzReadCurrentFile(uf,buf,size_buf);
	if (err<0) {
	    ap_log_rerror(APLOG_MARK,APLOG_ERR,0,r,
			  "(ziplook) error %d with zipfile in unzReadCurrentFile\n",err);
	    apr_table_addn(r->notes,"ziplook", "sent-gzip (read-error)");
	    break;
	}
	if(err > 0) {
	    ap_rwrite(buf, err, r);
	}
    } while (err>0);


    if (err != UNZ_OK) {
	unzCloseCurrentFile(uf); /* don't lose the error */       
	return err;
    }

    i = 0;
    gtable[i++] = (file_info->crc >> 0)  & 0xff;
    gtable[i++] = (file_info->crc >> 8)  & 0xff;
    gtable[i++] = (file_info->crc >> 16) & 0xff;
    gtable[i++] = (file_info->crc >> 24) & 0xff;
    
    gtable[i++] = (file_info->uncompressed_size >> 0)  & 0xff;
    gtable[i++] = (file_info->uncompressed_size >> 8)  & 0xff;
    gtable[i++] = (file_info->uncompressed_size >> 16) & 0xff;
    gtable[i++] = (file_info->uncompressed_size >> 24) & 0xff;
    
    ap_rwrite(gtable, i, r);

    err = unzCloseCurrentFile (uf);
    if (err!=UNZ_OK)
    {
	ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
			  "error %d with zipfile in unzCloseCurrentFile\n",err);
	apr_table_addn(r->notes,"ziplook", "sent-gzip (close-error)");
    }

    apr_table_addn(r->notes,"ziplook", "sent-gzip");

    return err;
}


/*
 * send_uncompressed is used when
 * 
 */


static int send_uncompressed(zipdata *zd)
{
    request_rec *r  = zd->r;
    unzFile      uf = zd->uf;
    unz_file_info *file_info = zd->file_info;
    int err=UNZ_OK;
    void* buf;
    uInt size_buf;
    int ratio=0;


    if (file_info->uncompressed_size>0)
	ratio = (file_info->compressed_size*100)/file_info->uncompressed_size;

    size_buf = WRITEBUFFERSIZE;
    buf = (void*) apr_palloc(r->pool, size_buf);
    if (buf==NULL)
    {
	r->content_type = "text/html"; 
	// ap_send_http_header(r);
        ap_rprintf(r,"Error allocating memory\n");

	apr_table_addn(r->notes,"ziplook", "sent-uncomp (memory-allocation-error)");

        return UNZ_INTERNALERROR;
    }

    err = unzOpenCurrentFile(uf);
    if (err!=UNZ_OK)
    {
	r->content_type = "text/html"; 
	// ap_send_http_header(r);
	ap_rprintf(r,"error %d with zipfile in unzOpenCurrentFile\n",err);

	apr_table_addn(r->notes,"ziplook", "sent-uncomp (open-file-error)");

	return -1;
    }

    r->content_type = get_content_type(r,zd->file);

    apr_table_setn(r->headers_out,
		  "Zip-info",
		  apr_psprintf(r->pool, 
			       "crc=%8.8lx, csize=%ld, usize=%ld, ratio=%d, method=%s", 
			       file_info->crc,
			       file_info->compressed_size,
			       file_info->uncompressed_size,
			       ratio,
			       get_string_method(file_info)));

    ap_update_mtime(r,get_time(r, file_info));

    if((err = ap_meets_conditions(r)) != OK) {
	r->status = err;
	ap_set_last_modified(r);
	// ap_send_http_header(r);

	apr_table_addn(r->notes,"ziplook", "sent-uncomp (not-modified)");

	return err;
    }

    ap_set_content_length(r, file_info->uncompressed_size);
    ap_set_last_modified(r);
    // ap_send_http_header(r);

    do {
	err = unzReadCurrentFile(uf,buf,size_buf);
	if (err<0) {
	    ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
			  "(ziplook) error %d with zipfile in unzReadCurrentFile\n",err);
	    apr_table_addn(r->notes,"ziplook", "sent-uncomp (read-error)");
	    break;
	}
	if(err > 0) {
	    ap_rwrite(buf, err, r);
	}
    } while (err>0);

    if (err==UNZ_OK)
    {
	err = unzCloseCurrentFile (uf);
	if (err!=UNZ_OK)
	{
	    ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, r,
			  "error %d with zipfile in unzCloseCurrentFile\n",err);
	    apr_table_addn(r->notes,"ziplook", "sent-uncomp (close-error)");
	} else {
	    apr_table_addn(r->notes,"ziplook", "sent-uncomp (ok)");
	}
    } else {
	unzCloseCurrentFile(uf); /* don't lose the error */ 
    }
    return err;
}





static int do_extract_currentfile(zipdata *zd)
{

    request_rec *r  = zd->r;
    unzFile      uf = zd->uf;
    int err=UNZ_OK;
    unz_file_info file_info;
    char filename_inzip[256];


    err = unzGetCurrentFileInfo(uf,&file_info,filename_inzip,sizeof(filename_inzip),NULL,0,NULL,0);

    if (err!=UNZ_OK)
    {
	r->content_type = "text/html"; 
	// ap_send_http_header(r);
	ap_rprintf(r,"error %d with zipfile in unzGetCurrentFileInfo\n",err);
	return err;
    }

    zd->file_info = &file_info;

    if((zd->accept_gzip) && (file_info.compression_method==Z_DEFLATED)) {
	err = send_compressed(zd);
    } else {
	err = send_uncompressed(zd);
    }

    return err;
}


static char *dir_index_names[] = {
  "index.html",
  "index.htm",
  NULL,
  NULL,
  NULL,
  NULL,
};

static int do_ext(zipdata *zd)
{
    int ret;
    char *path = zd->path;
    request_rec *r = zd->r;

    int len = strlen(path);

    if(path[len-1] != '/') {
	if (unzLocateFile(zd->uf,path+1,CASESENSITIVITY)!=UNZ_OK)
	    goto notfound;
    } else {
	int i;
        for(i=0; dir_index_names[i]; i++) {
	    path = apr_pstrcat(r->pool,zd->path,dir_index_names[i],NULL);
	    if (unzLocateFile(zd->uf,path+1,CASESENSITIVITY)==UNZ_OK) {
		goto found_it;
	    }
	}
	goto notfound;
    }
 found_it:
    zd->file = path;
    ret = do_extract_currentfile(zd);
    return OK;


 notfound:

    r->status = HTTP_NOT_FOUND;
    r->content_type = "text/html"; 
    // ap_send_http_header(r);
    ap_rprintf(r,"(ziplook) file not found (%s)", zd->path);
    return OK;
}



static apr_status_t unz_cleanup(void *voidp)
{
    unzFile uf= (void *) voidp;
    if(uf) {
	unzCloseCurrentFile(uf);
	unzClose(uf);
    }
    return APR_SUCCESS;
}


/* The sample content handler */
static int ziplook_handler(request_rec *r)
{
    zipdata *zd;
    char *zipfilename;
    unzFile uf;
    const char *enc;
    char *ext;
    char *sepa;
    apr_table_t *query;

    if (!strcmp(r->handler, "ziplook"))
	goto handle_this;

    if(!r->path_info)
	return DECLINED;

    if(r->finfo.filetype != APR_REG)
	return DECLINED;

    if(!(sepa = strrchr(r->filename,'/')))
	return DECLINED;
    if(!(ext = strrchr(r->filename,'.')))
	return DECLINED;
    if(sepa > ext)	
	return DECLINED;
    
    ext++;

    if(!strcasecmp(ext,"zip") || !strcasecmp(ext,"jar")) 
	goto handle_this;

    return DECLINED;
    
    
 handle_this:

    if(r->finfo.filetype != APR_REG) 
	return HTTP_NOT_FOUND;

    zipfilename = r->filename;
    uf = NULL;

    query = get_query(r);
    
    if(query && apr_table_get(query,"fetch")) {
	// this is fetch ... just use default handler
	return DECLINED;
    }

    ap_block_alarms();
    uf = unzOpen2(zipfilename, &memio);
    if(uf) apr_pool_cleanup_register(r->pool, uf, unz_cleanup, apr_pool_cleanup_null);
    ap_unblock_alarms();


    if(!uf)
	return HTTP_NOT_FOUND;

    zd = apr_pcalloc(r->pool, sizeof(zipdata));

    zd->r        = r;
    zd->query    = query;
    zd->filename = r->filename;
    zd->uf       = uf;


    if((enc  = apr_table_get(r->headers_in, "Accept-Encoding")) != NULL) {
#if TOO_COMPLICATED
	const char *ee = enc;
	while(ee && *ee) {



	    char *sp;
	    char *e = ap_getword(r->pool,&ee,',');
	    if(e == NULL)
		continue;

	    while(*e > 0 && *e <= 32) e++;

	    if((sp = strchr(e,';'))) *sp = 0;
	    if((sp = strchr(e,' '))) *sp = 0;

	    // we are only interested in gzip

	    if(!strcasecmp(e,"gzip") || !strcasecmp(e,"x-gzip")) {
		zd->accept_gzip = 1;
	    } 
	}
#else
	/* this is simple enough */
	if(strstr(enc,"gzip")) {
	    zd->accept_gzip = 1;
	}
#endif
    }

    if(!r->path_info || !*r->path_info) {
	zd->path = NULL;
	do_list(zd);
    } else {
	zd->path = r->path_info;
	do_ext(zd);
    }

    if(uf) {
	ap_block_alarms();
	unzClose(uf);
	apr_pool_cleanup_kill(r->pool, uf, unz_cleanup);
	ap_unblock_alarms();
    }

    return OK;
}

#ifndef ZIPLOOK_NO_VERSION_COMPONENT
static int ziplook_init_handler(apr_pool_t *p, apr_pool_t *plog, 
				apr_pool_t *ptemp, server_rec *s)
{
    ap_add_version_component(p, "mod_ziplook/" ZIPLOOK_VERSION_STR);
    return OK;
}
#endif

static void register_ziplook_hooks(apr_pool_t *p)
{
#ifndef ZIPLOOK_NO_VERSION_COMPONENT
    ap_hook_post_config(ziplook_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
#endif
    ap_hook_handler(ziplook_handler, NULL, NULL, APR_HOOK_MIDDLE);
}


module AP_MODULE_DECLARE_DATA ziplook_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                       /* dir config creater */
    NULL,                       /* dir merger --- default is to override */
    NULL,                       /* server config */
    NULL,                       /* merge server config */
    NULL,                       /* command table */
    register_ziplook_hooks      /* register_hooks */
};

